# Copyright 2016, Adrian Sampson.
# Copyright 2019, Christopher Larson.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

"""Prototyped version of `modify` whose values are defined with a template format string."""

from __future__ import division, absolute_import, print_function

from beets import library, ui
from beets.plugins import BeetsPlugin
from beets.ui import decargs, print_
from beets.ui.commands import _do_query, modify_parse_args, print_and_modify


class ModifyTmplPlugin(BeetsPlugin):
    def __init__(self):
        super(ModifyTmplPlugin, self).__init__()

    def commands(self):
        modify_cmd = ui.Subcommand(
            u'modifytmpl', help=u'change metadata fields, evaluating templates', aliases=(u'modt',)
        )
        modify_cmd.parser.add_option(
            u'-m', u'--move', action='store_true', dest='move',
            help=u"move files in the library directory"
        )
        modify_cmd.parser.add_option(
            u'-M', u'--nomove', action='store_false', dest='move',
            help=u"don't move files in library"
        )
        modify_cmd.parser.add_option(
            u'-w', u'--write', action='store_true', default=None,
            help=u"write new metadata to files' tags (default)"
        )
        modify_cmd.parser.add_option(
            u'-W', u'--nowrite', action='store_false', dest='write',
            help=u"don't write metadata (opposite of -w)"
        )
        modify_cmd.parser.add_album_option()
        modify_cmd.parser.add_format_option(target='item')
        modify_cmd.parser.add_option(
            u'-y', u'--yes', action='store_true',
            help=u'skip confirmation'
        )
        modify_cmd.func = modify_func
        return [modify_cmd]


def modify_func(lib, opts, args):
    query, mods, dels = modify_parse_args(decargs(args))
    if not mods and not dels:
        raise ui.UserError(u'no modifications specified')
    modify_items(lib, mods, dels, query, ui.should_write(opts.write),
                 ui.should_move(opts.move), opts.album, not opts.yes)


# Copied and tweaked from beets.ui.commands to apply a template
def modify_items(lib, mods, dels, query, write, move, album, confirm):
    """Modifies matching items according to user-specified assignments and
    deletions.

    `mods` is a dictionary of field and value pairse indicating
    assignments. `dels` is a list of fields to be deleted.
    """
    # Parse key=value specifications into a dictionary.
    model_cls = library.Album if album else library.Item

    # Get the items to modify.
    items, albums = _do_query(lib, query, album, False)
    objs = albums if album else items

    # Apply changes *temporarily*, preview them, and collect modified
    # objects.
    print_(u'Modifying {0} {1}s.'
           .format(len(objs), u'album' if album else u'item'))
    changed = []
    for obj in objs:
        # Changes from `modify` command here:
        obj_mods = {}
        for key, value in mods.items():
            value = obj.evaluate_template(value)
            obj_mods[key] = model_cls._parse(key, value)

        if print_and_modify(obj, obj_mods, dels) and obj not in changed:
            changed.append(obj)

    # Still something to do?
    if not changed:
        print_(u'No changes to make.')
        return

    # Confirm action.
    if confirm:
        if write and move:
            extra = u', move and write tags'
        elif write:
            extra = u' and write tags'
        elif move:
            extra = u' and move'
        else:
            extra = u''

        changed = ui.input_select_objects(
            u'Really modify%s' % extra, changed,
            lambda o: print_and_modify(o, mods, dels)
        )

    # Apply changes to database and files
    with lib.transaction():
        for obj in changed:
            obj.try_sync(write, move)



"""Run inline python code when beets events are fired."""

from __future__ import division, absolute_import, print_function

import ast
import confuse

from beets.plugins import BeetsPlugin


def _syntaxerror_offset(value, lineoffset):
    """Adjust the line number in a SyntaxError exception."""
    if lineoffset:
        msg, (efname, elineno, eoffset, badline) = value.args
        value.args = (msg, (efname, elineno + lineoffset, eoffset, badline))
        value.lineno = elineno + lineoffset


def compile_offset(source, filename='<string>', lineoffset=0):
    """Compile the python source and adjust its line numbers by lineoffset."""
    try:
        compiled = compile(source, filename, 'exec', ast.PyCF_ONLY_AST)
    except SyntaxError as exc:
        _syntaxerror_offset(exc, lineoffset)
        raise

    if lineoffset:
        ast.increment_lineno(compiled, lineoffset)

    return compile(compiled, filename, 'exec', dont_inherit=True)


def compile_func(source, name, argspec='', filename='<string>', lineoffset=0,
                 env=None):
    """Compile the python source, wrapped in a function definition."""
    # Adjust for 'def' line
    lineoffset -= 1

    code = source.rstrip().replace('\t', '    ')
    lines = ('    ' + line for line in code.split('\n'))
    code = '\n'.join(lines)
    defined = 'def {name}({argspec}):\n{body}'.format(name=name,
                                                      argspec=argspec,
                                                      body=code)
    compiled = compile_offset(defined, filename, lineoffset)

    if env is None:
        env = {}
    tmpenv = {}
    exec(compiled, env, tmpenv)
    return eval(name, env, tmpenv)


class InlineHookPlugin(BeetsPlugin):
    """Run inline python code when beets events are fired."""

    argspecs = {
        'after_write': 'item, path',
        'album_imported': 'lib, album',
        'albuminfo_received': 'info',
        'art_set': 'album',
        'before_item_moved': 'item, source',
        'cli_exit': 'lib',
        'database_change': 'lib, model',
        'import': 'lib, paths',
        'import_begin': 'session',
        'import_task_apply': 'session, task',
        'import_task_choice': 'session, task',
        'import_task_created': 'session, task',
        'import_task_files': 'session, task',
        'import_task_start': 'session, task',
        'item_copied': 'item, source',
        'item_hardlinked': 'item, source',
        'item_imported': 'lib, item',
        'item_linked': 'item, source',
        'item_moved': 'item, source',
        'item_removed': 'item',
        'library_opened': 'lib',
        'trackinfo_received': 'info',
        'write': 'item, path, tags',
    }

    def __init__(self):
        super(InlineHookPlugin, self).__init__()

        self.config.add({
            'hooks': []
        })

        inline_hooks = self.config['hooks'].get(list)

        for hook_index in range(len(inline_hooks)):
            hook = self.config['hooks'][hook_index]
            event = hook['event'].as_str()
            if event not in self.argspecs:
                raise confuse.ConfigError('inline_hook.hooks[{0}].event: `{1}` is not a handled event'.format(hook_index, event))
            handler = hook['handler'].as_str()
            # TODO: determine config value line number and use for lineoffset
            function = compile_func(handler, 'inline_hook_' + event, self.argspecs.get(event) or '')
            self.register_listener(event, function)



r"""
Path format function to apply the specified set of replacements to a path.

This can be used to apply replacements only to the alternative library created
by beets-alternatives, for example, while leaving the original library alone.

`replace_path` operates on a per-path-component basis, while `replace`
operates on the whole. The former is useful for replacing across a full path,
the latter on a field, as the latter can handle the case where the field
contains a path separator.

`sub` and `sub_path` are the same as `replace` and `replace_path`, but apply
a single specified replacement rather than referencing a set of replacements
in the config.

Usage:

  foo_replace:
    Foo: Bar

  bar_replace:
    '/.*': ''

  paths:
    default: '%replace_path{foo_replace,$album/$track - $artist - $title}'
    genre:foo: 'Foo/%replace{bar_replace,$album}/$track - $artist - $title'
    genre:bar: '%sub{$album,Foo,Bar}/$track - $artist - $title'
    singleton:bar: '%sub_path{[non-album tracks]/$track - $artist - $title,^\.}'
"""

from __future__ import division, absolute_import, print_function

import functools
import re

import confuse

from beets import config, util
from beets.ui import UserError
from beets.plugins import BeetsPlugin


class ReplacePlugin(BeetsPlugin):
    def __init__(self):
        super(ReplacePlugin, self).__init__()
        self.template_funcs['replace'] = self.replace
        self.template_funcs['replace_path'] = self.replace_path
        self.template_funcs['sub'] = self.sub
        self.template_funcs['sub_path'] = self.sub_path

    def replace(self, config_path, string):
        replacements = get_replacements(config_path)
        for regex, repl in replacements:
            string = regex.sub(repl, string)
        return string

    def replace_path(self, config_path, path):
        replacements = get_replacements(config_path)
        return util.sanitize_path(path, replacements)

    def sub(self, string, pattern, repl=''):
        try:
            return re.compile(pattern).sub(repl, string)
        except re.error as exc:
            raise UserError(u'%sub: error compiling regex `{0}`: {1}'.format(pattern, str(exc)))

    def sub_path(self, path, pattern, repl=''):
        try:
            return util.sanitize_path(path, [(re.compile(pattern), repl)])
        except re.error as exc:
            raise UserError(u'%sub_path: error compiling regex `{0}`: {1}'.format(pattern, str(exc)))


# Copied with tweak from beets itself
def get_replacements(config_path):
    replacements = []
    lookup = functools.reduce(lambda x, y: x[y], config_path.split("."), config)
    for pattern, repl in lookup.get(dict).items():
        try:
            replacements.append((re.compile(pattern), repl or ''))
        except re.error as exc:
            raise confuse.ConfigError(u'{0}: error compiling regex `{1}`: {2}'.format(config_path, pattern, str(exc)))
    return replacements



"""Run 'beet picard' on the paths to be imported.

This is useful when you want to use MusicBrainz Picard to examine or manipulate
the tags of the files being imported/reimported. Spawning picard will differ
between platforms, so we assume a `beet-picard` alias/script is available to do
the work.
"""

from __future__ import division, absolute_import, print_function

import subprocess

from beets import ui, util
from beets.plugins import BeetsPlugin
from beets.ui.commands import PromptChoice


class ImportPicardPlugin(BeetsPlugin):
    def __init__(self):
        super(ImportPicardPlugin, self).__init__()

        self.register_listener('before_choose_candidate',
                               self.before_choose_candidate_listener)

    def before_choose_candidate_listener(self, session, task):
        return [PromptChoice('p', 'Picard', self.run_picard)]

    def run_picard(self, session, task):
        paths = [util.syspath(item.path) for item in task.items]
        argv = ['beet', 'picard'] + paths
        try:
            subprocess.check_call(argv)
        except subprocess.CalledProcessError as exc:
            raise ui.UserError(str(exc))
        return None



"""Require the 'source' field be set, to organize the library by where the music was obtained.

Ex. iTunes, Amazon, etc.
"""

from __future__ import division, absolute_import, print_function

from beets import config, ui
from beets.plugins import BeetsPlugin


class MusicSourcePlugin(BeetsPlugin):
    def __init__(self):
        super(MusicSourcePlugin, self).__init__()
        self.register_listener('import_task_created', self.import_task_created)

    def import_task_created(self, session, task):
        try:
            reimporting = config['import']['library'].get()
        except Exception:
            reimporting = False

        if not reimporting:
            set_fields = config['import']['set_fields']
            if 'source' not in set_fields.keys():
                raise ui.UserError('No source specified, please include --set=source=SOURCE')



"""Specify fields from set_fields to also be applied to skipped items on reimport.

This is a bit special-purpose. I have one single use case in mind for this, resuming
a *reimport*, ex:

    reimportskipfields:
      set_fields: reimported

    beet import --set=reimported=1 '^reimported:1' mb_albumid::'^$'

This ensures that the field is set regardless of whether the album/item was matched
or skipped, so both are ignored in subsequent import passes. When I add a new autotag
plugin to my configuration, I just use modify to wipe the reimported fields and start
with a new import pass.
"""

from __future__ import division, absolute_import, print_function

from beets import config, dbcore
from beets.util import displayable_path
from beets.plugins import BeetsPlugin


class ReimportSkippedFieldsPlugin(BeetsPlugin):
    def __init__(self):
        super(ReimportSkippedFieldsPlugin, self).__init__()

        self.config.add({
            'set_fields': '',
        })
        self.set_fields = self.config['set_fields'].as_str_seq()

        self.register_listener('import_task_choice', self.import_task_choice)

    def import_task_choice(self, session, task):
        try:
            reimporting = config['import']['library'].get()
        except Exception:
            pass
        else:
            if reimporting and task.skip:
                self.mark_items(session, task, task.items)

    def mark_items(self, session, task, items):
        fields = self.fields()
        marked = set()
        for item in items:
            if not item.id:
                # Same logic as used by `record_replaced` in `ImportTask` in beets.importer
                old_items = list(session.lib.items(
                    dbcore.query.BytesQuery('path', item.path)
                ))
                item = old_items[0]

            if task.is_album:
                album = item.get_album()
                if album:
                    for field, value in fields:
                        self._log.debug(u'Set field {1}={2} for {0}', displayable_path(album.path), field, value)
                        album[field] = value
                    marked.add(album)
                break
            else:
                for field, value in fields:
                    self._log.debug(u'Set field {1}={2} for {0}', displayable_path(item.path), field, value)
                    item[field] = value
                marked.add(item)

        with session.lib.transaction():
            for item in marked:
                item.store()

    def fields(self):
        for field, view in config['import']['set_fields'].items():
            if field in self.set_fields:
                value = view.get()
                yield field, value



"""Apply modifications on album/item import."""

from __future__ import division, absolute_import, print_function

from beets import dbcore, ui, util
from beets.library import parse_query_string, Item, Album
from beets.plugins import BeetsPlugin
from beets.ui import decargs
from beets.ui.commands import modify_parse_args, modify_items
from beets.util import shlex_split


class ModifyOnImport(BeetsPlugin):
    def __init__(self):
        super(ModifyOnImport, self).__init__()

        self.config.add({
            'modify_album': {},
            'modify_album_items': {},
            'modify_singleton': {},
        })

        self.import_stages = [self.imported]
        self.register_listener('import_begin', self.import_begin)

    def import_begin(self, session):
        self.should_write = ui.should_write()
        self.should_move = ui.should_move()

        for name, model_cls in [('album', Album), ('singleton', Item)]:
            modifies = self.get_modifies(self.config['modify_' + name].items(), model_cls)
            setattr(self, name + '_modifies', modifies)

        self.album_item_modifies = []
        for albumquery, itemmodifies in self.config['modify_album_items'].items():
            albumdbquery, _ = parse_query_string(util.as_string(albumquery), Album)
            modifies = self.get_modifies(itemmodifies.items(), Item)
            self.album_item_modifies.append((albumdbquery, modifies))

    def get_modifies(self, items, model_cls):
        modifies = []
        for query, modify in items:
            modify = modify.as_str()
            _, mods, dels = self.parse_modify(modify, model_cls)
            dbquery, _ = parse_query_string(util.as_string(query), model_cls)
            modifies.append((dbquery, mods, dels))
        return modifies

    def parse_modify(self, modify, model_cls):
        modify = util.as_string(modify)
        args = shlex_split(modify)
        query, mods, dels = modify_parse_args(decargs(args))
        query = util.as_string(' '.join(query))
        dbquery, _ = parse_query_string(query, model_cls)
        return dbquery, mods, dels

    def imported(self, session, task):
        if task.is_album:
            modifies = self.album_modifies
        else:
            modifies = self.singleton_modifies

        objs = task.imported_items()
        self.modify_objs(session.lib, objs, modifies, task.is_album)
        if task.is_album:
            for albumdbquery, modifies in self.album_item_modifies:
                if albumdbquery.match(task.album):
                    self.modify_objs(session.lib, task.album.items(), modifies, False)

    def modify_objs(self, lib, objs, modifies, is_album):
        for obj in objs:
            all_mods, all_dels = {}, {}
            for dbquery, mods, dels in modifies:
                if dbquery.match(obj):
                    all_mods.update(mods)
                    all_dels.update(dels)
            if all_mods or all_dels:
                idquery = dbcore.MatchQuery('id', obj.id)
                modify_items(lib, all_mods, all_dels, idquery, self.should_write, self.should_move, is_album, False)



"""Support for beets command aliases, not unlike git.

By default, also checks $PATH for beet-* and makes those available as well.

Example:

    alias:
      from_path: yes # Default
      aliases:
        singletons: ls singleton:true
        external-cmd-test: '!echo'
        sh-c-test: '!sh -c "echo foo bar arg1:$1, arg2:$2" sh-c-test'
        with-help-text:
          command: ls -a
          help: do something or other
"""

from __future__ import division, absolute_import, print_function

import confuse
import glob
import optparse
import os
import six
import subprocess
import sys

from beets import plugins
from beets import ui
from beets.plugins import BeetsPlugin
from beets.ui import Subcommand, print_
from beets.ui.commands import default_commands
from beets.util import shlex_split

if sys.version_info >= (3, 3):
    from collections import abc
else:
    import collections as abc


class NoOpOptionParser(optparse.OptionParser):
    def parse_args(self, args=None, namespace=None):
        if args is None:
            args = sys.argv[1:]
        return [], args


class AliasCommand(Subcommand):
    def __init__(self, alias, command, log=None, help=None):
        super(AliasCommand, self).__init__(alias, help=help or command, parser=NoOpOptionParser(add_help_option=False, description=help or command))

        self.alias = alias
        self.log = log
        self.command = command

    def func(self, lib, opts, args=None):
        if args is None:
            args = []

        split_command = shlex_split(self.command)
        command = split_command[0]
        args = split_command[1:] + args
        if command.startswith('!'):
            command = command[1:]
            argv = [command] + args

            if self.log:
                self.log.debug('Running {}', subprocess.list2cmdline(argv))

            try:
                return subprocess.check_call(argv)
            except subprocess.CalledProcessError as exc:
                if self.log:
                    self.log.debug(u'command `{0}` failed with {1}', subprocess.list2cmdline(argv), exc.returncode)
                plugins.send('cli_exit', lib=lib)
                lib._close()
                sys.exit(exc.returncode)
        else:
            argv = [command] + args
            cmdname = argv[0]

            subcommands = list(default_commands)
            subcommands.extend(plugins.commands())
            for subcommand in subcommands:
                if cmdname == subcommand.name or cmdname in subcommand.aliases:
                    break
            else:
                raise ui.UserError(u"unknown command '{0}'".format(cmdname))

            if self.log:
                self.log.debug('Running {}', subprocess.list2cmdline(argv))

            suboptions, subargs = subcommand.parse_args(argv[1:])
            return subcommand.func(lib, suboptions, subargs)


class AliasPlugin(BeetsPlugin):
    """Support for beets command aliases, not unlike git."""

    def __init__(self):
        super(AliasPlugin, self).__init__()

        self.config.add({
            'from_path': True,
            'aliases': {},
        })

    def get_command(self, alias, command, help=None):
        """Create a Subcommand instance for the specified alias."""
        return AliasCommand(alias, command, log=self._log, help=help)

    def get_path_commands(self):
        """Create subcommands for beet-* scripts in $PATH."""
        for path in os.getenv('PATH', '').split(':'):
            cmds = glob.glob(os.path.join(path, 'beet-*'))
            for cmd in cmds:
                if os.access(cmd, os.X_OK):
                    command = os.path.basename(cmd)
                    alias = command[5:]
                    yield (alias, self.get_command(alias, '!' + command, u'Run external command `{0}`'.format(command)))

    def cmd_alias(self, lib, opts, args, commands):
        """Print the available alias commands."""
        for alias, command in sorted(commands.items()):
            print_(u'{0}: {1}'.format(alias, command))

    def commands(self):
        """Add the alias commands."""
        if self.config['from_path'].get(bool):
            commands = dict(self.get_path_commands())
        else:
            commands = {}

        for alias in self.config['aliases'].keys():
            if alias in commands:
                raise confuse.ConfigError(u'alias.aliases.{0} was specified multiple times'.format(alias))

            command = self.config['aliases'][alias].get()
            if isinstance(command, six.text_type):
                commands[alias] = self.get_command(alias, command)
            elif isinstance(command, abc.Mapping):
                command_text = command.get('command')
                if not command_text:
                    raise confuse.ConfigError(u'alias.aliases.{0}.command not found'.format(alias))
                help_text = command.get('help', command_text)
                commands[alias] = self.get_command(alias, command_text, help_text)
            else:
                raise confuse.ConfigError(u'alias.aliases.{0} must be a string or single-element mapping'.format(alias))

        if 'alias' in commands:
            raise ui.UserError(u'alias `alias` is reserved for the alias plugin')

        alias = Subcommand('alias',
                           help=u'Print the available alias commands.')
        alias_commands = dict((a, c.command) for a, c in commands.items())
        alias.func = lambda lib, opts, args: self.cmd_alias(lib, opts, args, alias_commands)
        commands['alias'] = alias
        return commands.values()



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['beetsplug']

package_data = \
{'': ['*']}

install_requires = \
['beets>=1.4,<2.0', 'confuse>=1.0.0,<2.0.0']

setup_kwargs = {
    'name': 'beets-kergoth',
    'version': '0.3.2',
    'description': "Various personal beets plugins that don't yet have a home elsewhere.",
    'long_description': "beets-kergoth\n=============\n\nThis is my personal collection of beets_ plugins that are not currently being\ndistributed separately.\n\nGenerally useful\n----------------\n\nThese plugins I consider to be useful in general and polished enough that I\nintend to either submit them upstream to beets or split off into separate\nprojects / python packages.\n\n- The `Alias Plugin`_ lets you define command aliases, much like git, and also\n  makes available any beet-prefixed commands in your ``PATH`` as well.\n- The `Import Inspect Plugin`_ lets you inspect the metadata changes to be\n  applied when importing.\n- The `Modify On Import Plugin`_ lets you define metadata changes to occur\n  on import when the specified queries match.\n- The `Replace Format Plugin`_ defines format functions for search/replace of\n  text, both single replacements and applications of a set of replacements\n  like the built in ``replace`` configuration option.\n- The `Saved Formats Plugin`_ lets you define saved, named format strings by\n  storing them in fields for later reference.\n- The `Saved Queries Plugin`_ lets you save queries by name for later use.\n\nSpecial Purpose\n---------------\n\nThese plugins are useful to me, but probably not to anyone else.\n\n- The ``musicsource`` plugin just forces me to define a ``source`` field when\n  importing new music, as this is how I organize my library. ``source`` is\n  where the music was acquired, e.g. Amazon, iTunes, Google, etc.\n- The ``picard`` plugin lets me launch MusicBrainz Picard with the items I'm\n  importing, either to use it to tag rather than beets, or as an inspection\n  tool to examine its metadata. This assumes a ``picard`` beets command\n  exists, as this is how it runs it.\n- The ``reimportskipfields`` plugin lets you specify fields from set_fields\n  to also be applied to skipped items on reimport. I use this to facilitate\n  resuming from a reimport, as it lets me apply a ``reimported`` field to\n  anything I tried to reimport, whether I was able to match it to a candidate\n  or not.\n\nPrototype / Proof of Concept plugins\n------------------------------------\n\nThese plugins were thrown together in experimentation. Some may be worth\ncleaning up, but others are useful pretty rarely.\n\n- The ``abcalc`` plugin performs the same calculation as absubmit, but runs it\n  only against non-MusicBrainz tracks, and stores the low level values rather\n  than submitting them. I've used this to populate the acoustic fingerprint\n  tags for use by external de-duplication scripts.\n- The ``existingqueries`` plugin adds a couple queries that originated in the\n  beets source.\n- The ``hasart`` plugin adds a query to check for embedded album art.\n- The ``inlinehook`` plugin lets you define hooks inline in config.yaml with\n  python, much the way ``inline`` does for fields.\n- The ``last_import`` plugin keeps track of the most recently imported items\n  with a flexible field.\n- The ``modifytmpl`` plugin lets you define fields using templates / format\n  strings. This will be going upstream into the main ``modify`` command.\n- The ``otherqueries`` plugin defines other random queries of questionable\n  usefulness at this time.\n- The ``spotifyexplicit`` plugin uses the ``spotify`` plugin to look up items\n  from my library, determines if Spotify considers these items as explicit\n  tracks (Parental Advisory), and prints them if so. I use this to set an\n  ``advisory`` field on my tracks.\n\n\n.. _beets: http://beets.io/\n.. _Alias Plugin: docs/alias.rst\n.. _Import Inspect Plugin: docs/importinspect.rst\n.. _Modify On Import Plugin: docs/modifyonimport.rst\n.. _Replace Format Plugin: docs/replaceformat.rst\n.. _Saved Formats Plugin: docs/savedformats.rst\n.. _Saved Queries Plugin: docs/savedqueries.rst\n",
    'author': 'Christopher Larson',
    'author_email': 'kergoth@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/kergoth/beets-kergoth',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.8.*',
}


setup(**setup_kwargs)



"""Support saved, named queries."""

from __future__ import division, absolute_import, print_function

from beets import config
from beets.library import parse_query_string, Item, Album
from beets.plugins import BeetsPlugin

from beets.dbcore import types
from beets.dbcore import Query


class SavedQuery(Query):
    model_name = 'invalid'

    def __init__(self, name):
        self.name = name
        queries = config[self.model_name + '_queries']
        self.query_string = queries[name].as_str()
        self.query, _ = parse_query_string(self.query_string, self.model)

    def clause(self):
        return self.query.clause()

    def match(self, item):
        return self.query.match(item)


class ItemSavedQuery(SavedQuery):
    model_name = 'item'
    model = Item


class AlbumSavedQuery(SavedQuery):
    model_name = 'album'
    model = Album


class SavedQueriesPlugin(BeetsPlugin):
    """Support saved, named queries.

    Also support creation of Item/Album boolean fields for the query. Do not
    query the fields, as the performance is not great. Use the `query:NAME`
    query syntax for queries, and use the fields for inspection or a format
    string.
    """

    def __init__(self):
        super(SavedQueriesPlugin, self).__init__()

        self.config.add({
            'add_fields': True,
        })

        config.add({
            'item_queries': {},
            'album_queries': {},
        })

        self._log.debug('adding named item query `query`')
        self._log.debug('adding named album query `album_query`')
        self.item_queries = {'query': ItemSavedQuery}
        self.album_queries = {'album_query': AlbumSavedQuery}
        if self.config['add_fields'].get():
            self.item_types = {}
            self.template_fields = {}
            for name in config['item_queries'].keys():
                self._log.debug('adding item field {}', name)
                self.item_types[name] = types.BOOLEAN
                self.template_fields[name] = lambda item, name=name: ItemSavedQuery(name).match(item)

            self.album_types = {}
            self.album_template_fields = {}
            for name in config['album_queries'].keys():
                self._log.debug('adding album field {}', name)
                self.album_types[name] = types.BOOLEAN
                self.album_template_fields[name] = lambda album, name=name: AlbumSavedQuery(name).match(album)



from beets import ui
from beets.plugins import BeetsPlugin, find_plugins


class SpotifyExplicitPlugin(BeetsPlugin):
    def commands(self):
        for plugin in find_plugins():
            if plugin.name == 'spotify':
                self.spotify = plugin
                break
        else:
            raise ui.UserError('spotify plugin is required')

        def explicits(lib, opts, args):
            results = self.spotify._match_library_tracks(lib, ui.decargs(args))
            if results:
                for track in results:
                    if track['explicit']:
                        title = track['name']
                        album = track['album']['name']
                        artist = track['artists'][0]['name']
                        tracknum = track['track_number']
                        url = track['external_urls']['spotify']
                        print('{} - {} - {} - {} - {}'.format(album, tracknum, artist, title, url))

        explicit_cmd = ui.Subcommand(
            'spotify-explicit', help=u''
        )
        explicit_cmd.parser.add_all_common_options()
        explicit_cmd.func = explicits
        return [explicit_cmd]
